import torch
import random
import numpy as np
import skimage as skimg
import cv2
import math
from PIL import Image, ImageOps, ImageFilter
import PIL.ImageEnhance as ImageEnhance
import time

class GaussianNoise(object):
    def __init__(self,mean=0,std=0):
        self.mean = mean
        self.std = std
    def __call__(self,sample):
        key_list = sample.keys()
        for key in key_list:
            if 'image' in key:
                img = sample[key]
                sample[key] = skimg.util.random_noise(img,mode='gaussian',mean=self.mean,var=(self.std)**2)

        return sample

class Normalize(object):
    """Normalize a tensor image with mean and standard deviation.
    Args:
        mean (tuple): means for each channel.
        std (tuple): standard deviations for each channel.
    """
    def __init__(self, mean=[123.675, 116.280, 103.530], std=[58.395, 57.120, 57.375]):
        if isinstance(std, (int, float)):
            std = [std, std, std]
        else:
            self.mean = np.array(mean)
            self.std = np.array(std)

    def __call__(self, sample):
        key_list = sample.keys()
        for key in key_list:
            if 'image' in key:
                img = sample[key].astype(np.float64)
                img -= self.mean
                img /= self.std
                sample[key] = img
        return sample

class ToTensor(object):
    """Convert ndarrays in sample to Tensors."""

    def __call__(self, sample):
        key_list = list(sample.keys())
        for key in key_list:
        # swap color axis because
        # numpy image: H x W x C
        # torch image: C X H X W
            if 'image' == key:
                img = sample[key]
                if len(img.shape) == 2:
                    img = np.expand_dims(img, axis=0).copy()
                else:
                    img = img.transpose((2, 0, 1)).copy()
                sample[key] = torch.from_numpy(img).float()
            elif 'audio' == key:
                aud = sample[key]
                # aud = aud.reshape((1,-1,1))
                sample[key] = torch.from_numpy(aud).float()
            elif 'label' in key:
                label = sample[key]
                # sample[key] = torch.from_numpy(label).long()
                sample[key] = torch.tensor(label).long()
        return sample


class RandomHorizontalFlip(object):
    def __call__(self, sample):
        if np.random.rand() < 0.5:
            key_list = sample.keys()
            for key in key_list:
                if 'image' not in key:
                    continue
                image = sample[key]
                image_flip = np.flip(image, axis=1)
                sample[key] = image_flip
        return sample


class RandomRotate(object):
    """Randomly rotate image"""
    def __init__(self, angle_r,image_value=127,is_continuous=True):
        self.angle_r = angle_r
        self.seg_interpolation = cv2.INTER_LINEAR if is_continuous else cv2.INTER_NEAREST
        self.IMAGE_VALUE = image_value

    def __call__(self, sample):
        # t0 = time.time()
        if np.random.rand() < 0.5:
            return sample
        rand_angle = np.random.randint(-self.angle_r, self.angle_r) if self.angle_r != 0 else 0
        PI = 3.141592653
        Hangle = rand_angle*PI / 180
        Hcos = math.cos(Hangle)
        Hsin = math.sin(Hangle)
        key_list = sample.keys()
        for key in key_list:
            if 'image' not in key:
                continue
            image = sample[key]
            imgsize = image.shape
            srcWidth = imgsize[1]
            srcHeight = imgsize[0]
            x = [0,0,0,0]
            y = [0,0,0,0]
            x1 = [0,0,0,0]
            y1 = [0,0,0,0]
            x[0] = -(srcWidth - 1) / 2
            x[1] = -x[0]
            x[2] = -x[0]
            x[3] = x[0]
            y[0] = -(srcHeight - 1) / 2
            y[1] = y[0]
            y[2] = -y[0]
            y[3] = -y[0]
            for i in range(4):
                x1[i] = int(x[i] * Hcos + y[i] * Hsin + 0.5)
                y1[i] = int(-x[i] * Hsin + y[i] * Hcos + 0.5)
            if (abs(y1[2] - y1[0]) > abs(y1[3] - y1[1])):
                Height = abs(y1[2] - y1[0])
                Width = abs(x1[3] - x1[1])
            else:
                Height = abs(y1[3] - y1[1])
                Width = abs(x1[2] - x1[0])
            row, col = image.shape[:2]
            m = cv2.getRotationMatrix2D(center=(col/2, row/2), angle=rand_angle, scale=1)
            new_image = cv2.warpAffine(image, m, (Width,Height), flags=cv2.INTER_LINEAR if 'image' in key else self.seg_interpolation, 
                borderValue=self.IMAGE_VALUE if 'image' in key else self.MASK_VALUE)
            sample[key] = new_image
        
        # t1 = time.time()
        # print('Rotate: %.4f'%(1000*(t1 - t0)))
        return sample

class Resize(object):
    def __init__(self,output_size,is_continuous=False,label_size=None):
        assert isinstance(output_size, (tuple,list))
        if len(output_size) == 1:
            self.output_size = (output_size[0],output_size[0])
        else:
            self.output_size = tuple(output_size)
        self.seg_interpolation = cv2.INTER_LINEAR if is_continuous else cv2.INTER_NEAREST
    
    def __call__(self,sample):
        # t0 = time.time()
        if not 'image' in sample.keys():
            return sample

        key_list = sample.keys()
        for key in key_list:
            if key != 'image':
                continue
            img = sample[key]
            h, w = img.shape[:2]
            img = cv2.resize(img, dsize=self.output_size, interpolation=cv2.INTER_LINEAR if 'image' in key else self.seg_interpolation)
            sample[key] = img

        # t1 = time.time()
        # print('Resize: %.4f'%(1000*(t1 - t0)))
        return sample

class RandomScale(object):
    def __init__(self,rand_resize,is_continuous=False):
        self.rand_resize = rand_resize
        self.seg_interpolation = cv2.INTER_LINEAR if is_continuous else cv2.INTER_NEAREST

    def __call__(self, sample):
        rand_scale = random.uniform(self.rand_resize[0], self.rand_resize[1])
        key_list = sample.keys()
        for key in key_list:
            if 'image' not in key and 'label' not in key:
                continue
            img = sample[key]
            img = cv2.resize(img, None, fx=rand_scale, fy=rand_scale, interpolation=cv2.INTER_LINEAR if 'image' in key else self.seg_interpolation)
            sample[key] = img

        if 'bbox' in key_list:
            x0, y0, x1, y1 = sample['bbox']
            x0 *= rand_scale
            x1 *= rand_scale
            y0 *= rand_scale
            y1 *= rand_scale
            sample['bbox'] = np.array([x0, y0, x1, y1])

        return sample

class RandomCrop(object):
    def __init__(self,crop_size, mask_value=0, image_value=127):
        assert isinstance(crop_size, (tuple,list))
        if len(crop_size) == 1:
            self.crop_size = (crop_size[0],crop_size[0])
        else:
            self.crop_size = crop_size
        self.MASK_VALUE = mask_value
        self.IMAGE_VALUE = image_value

    def __call__(self,sample):
        rand_pad = random.uniform(0, 1)
        key_list = sample.keys()
        for key in key_list:
            if 'image' not in key:
                continue
            img = sample[key]
            h,w = img.shape[:2]
            new_h,new_w = self.crop_size
            pad_w = new_w - w
            pad_h = new_h - h
            w_begin = max(0,-pad_w)
            h_begin = max(0,-pad_h)
            pad_w = max(0,pad_w)
            pad_h = max(0,pad_h)
            w_begin = int(w_begin * rand_pad)
            h_begin = int(h_begin * rand_pad)
            w_end = w_begin + min(w,new_w)
            h_end = h_begin + min(h,new_h)
            shape = list(img.shape)
            shape[0] = new_h
            shape[1] = new_w
            new_img = np.zeros(shape,dtype=np.float64)
            new_img.fill(self.IMAGE_VALUE)
            new_img[pad_h//2:min(h,new_h)+pad_h//2,pad_w//2:min(w,new_w)+pad_w//2] = img[h_begin:h_end,w_begin:w_end]
            sample[key] = new_img

        return sample

class RandomShift(object):
    def __init__(self,shift_pixel,mask_value,image_value):
        self.shift_pixel = shift_pixel
        self.MASK_VALUE = mask_value
        self.IMAGE_VALUE = image_value

    def __call__(self,sample):
        rand_x = int((random.uniform(0, 1)-0.5)*2*self.shift_pixel)
        rand_y = int((random.uniform(0, 1)-0.5)*2*self.shift_pixel)
        key_list = sample.keys()
        mean = sample['mean']
        for key in key_list:
            if 'image' not in key and 'label' not in key:
                continue
            img = sample[key]
            h,w = img.shape[:2]
            new_x0 = max(0,rand_x)
            new_y0 = max(0,rand_y)
            new_x1 = w + min(0, rand_x)
            new_y1 = h + min(0, rand_y)
            x0 = max(0, -rand_x)
            y0 = max(0, -rand_y)
            x1 = x0 + new_x1 - new_x0
            y1 = y0 + new_y1 - new_y0
            new_img = np.zeros_like(img, dtype=np.float)
            if 'image' in key:
                # if 'pad' in key_list:
                #     self.IMAGE_VALUE = sample['pad']
                new_img[...,0].fill(mean[0])
                new_img[...,1].fill(mean[1])
                new_img[...,2].fill(mean[2])
                # else:
                #     new_img.fill(self.IMAGE_VALUE)
            elif 'label' in key:
                new_img.fill(self.MASK_VALUE)
            new_img[new_y0:new_y1, new_x0:new_x1] = img[y0:y1, x0:x1]
            sample[key] = new_img
        
        if 'bbox' in key_list:
            x0, y0, x1, y1 = sample['bbox']
            x0 += rand_x
            x1 += rand_x
            y0 += rand_y
            y1 += rand_y
            sample['bbox'] = np.array([x0, y0, x1, y1])
        
        return sample

class CenterPad(object):
    def __init__(self, pad_size,mask_value,image_value):
        assert isinstance(pad_size, (tuple,list))
        if len(pad_size) == 1:
            self.pad_size = (pad_size[0],pad_size[0])
        else:
            self.pad_size = pad_size
        self.MASK_VALUE = mask_value
        self.IMAGE_VALUE = image_value

    def __call__(self,sample):
        key_list = sample.keys()
        for key in key_list:
            if 'image' not in key and 'label' not in key:
                continue
            img = sample[key]
            h,w = img.shape[:2]
            new_h,new_w = self.pad_size
            pad_w = new_w - w
            pad_h = new_h - h
            w_begin = max(0,-pad_w)
            h_begin = max(0,-pad_h)
            pad_w = max(0,pad_w)
            pad_h = max(0,pad_h)
            w_begin = int(w_begin * 0.5)
            h_begin = int(h_begin * 0.5)
            w_end = w_begin + min(w,new_w)
            h_end = h_begin + min(h,new_h)
            shape = list(img.shape)
            shape[0] = new_h
            shape[1] = new_w
            new_img = np.zeros(shape,dtype=np.float)
            if 'image' in key:
                new_img.fill(self.IMAGE_VALUE)
            elif 'label' in key:
                new_img.fill(self.MASK_VALUE)
            new_img[pad_h//2:min(h,new_h)+pad_h//2,pad_w//2:min(w,new_w)+pad_w//2] = img[h_begin:h_end,w_begin:w_end]
            sample[key] = new_img
        return sample

class CenterCrop(object):
    def __init__(self,crop_size,mask_value,image_value):
        assert isinstance(crop_size, (int, tuple,list))
        if isinstance(crop_size, int):
            self.crop_size = (crop_size,crop_size)
        elif isinstance(crop_size,list) and len(crop_size) == 1:
            self.crop_size = (crop_size[0],crop_size[0])
        else:
            self.crop_size = crop_size
        self.MASK_VALUE = mask_value
        self.IMAGE_VALUE = image_value
    
    def __call__(self,sample):
        key_list = sample.keys()
        mean = sample['mean']
        for key in key_list:
            if 'image' in key or 'label' in key:
                img = sample[key]
                h,w = img.shape[:2]
                new_h,new_w = self.crop_size
                
                if 'label' in key:
                    new_h = int(new_h * sample.get('resize_scale',1))
                    new_w = int(new_w * sample.get('resize_scale',1))
                pad_w = new_w - w
                pad_h = new_h - h
                w_begin = max(0,-pad_w)
                h_begin = max(0,-pad_h)
                pad_w = max(0,pad_w)
                pad_h = max(0,pad_h)
                w_begin = int(w_begin * 0.5)
                h_begin = int(h_begin * 0.5)
                w_end = w_begin + min(w,new_w)
                h_end = h_begin + min(h,new_h)
                shape = list(img.shape)
                shape[0] = new_h
                shape[1] = new_w
                new_img = np.zeros(shape,dtype=np.float)
                if 'image' in key:
                    new_img[...,0].fill(mean[0])
                    new_img[...,1].fill(mean[1])
                    new_img[...,2].fill(mean[2])
                    # new_img.fill(self.MASK_VALUE)
                elif 'label' in key:
                    new_img.fill(self.MASK_VALUE)
                new_img[pad_h//2:min(h,new_h)+pad_h//2,pad_w//2:min(w,new_w)+pad_w//2] = img[h_begin:h_end,w_begin:w_end]
                sample[key] = new_img

        if 'bbox' in key_list:
            x0, y0, x1, y1 = sample['bbox']
            x0 += pad_w//2 - w_begin
            x1 += pad_w//2 - w_begin
            y0 += pad_h//2 - h_begin
            y1 += pad_h//2 - h_begin
            sample['bbox'] = np.array([x0, y0, x1, y1])
                
        return sample

class BorderPad(object):
    def __init__(self,mask_value,image_value,size):
        self.MASK_VALUE = mask_value
        self.IMAGE_VALUE = image_value
        self.size = size
    
    def __call__(self, sample):
        key_list = sample.keys()
        mean = sample['mean']
        for key in key_list:
            if 'image' in key or 'label' in key:
                img = sample[key]
                h, w = img.shape[:2]
                if h >= self.size and w >= self.size:
                    return sample

                size = max(self.size,h,w)

                pad_w = (size - w)//2
                pad_h = (size - h)//2
                shape = list(img.shape)
                shape[0] = size
                shape[1] = size
                new_img = np.zeros(shape,dtype=np.float)
                if 'image' in key:
                    new_img[...,0].fill(mean[0])
                    new_img[...,1].fill(mean[1])
                    new_img[...,2].fill(mean[2])
                    # new_img.fill(self.MASK_VALUE)
                elif 'label' in key:
                    new_img.fill(self.MASK_VALUE)
                new_img[pad_h//2:h+pad_h//2,pad_w//2:w+pad_w//2] = img
                sample[key] = new_img

        if 'bbox' in key_list:
            x0, y0, x1, y1 = sample['bbox']
            x0 += pad_w//2
            x1 += pad_w//2
            y0 += pad_h//2
            y1 += pad_h//2
            sample['bbox'] = np.array([x0, y0, x1, y1])
                
        return sample

class CenterCropFromBBox(object):
    def __init__(self,k=1.5):
        self.k = k

    def __call__(self, sample):
        key_list = sample.keys()
        if not 'bbox' in key_list:
            return sample

        x0,y0,x1,y1 = sample['bbox']
        cx = (x0+x1)/2
        cy = (y0+y1)/2
        w = (x1 - x0)*self.k
        h = (y1 - y0)*self.k


        for key in key_list:
            if 'image' in key or 'label' in key:
                img = sample[key]
                H,W = img.shape[:2]

                x0 = max(0, int(cx - w/2))
                y0 = max(0, int(cy - h/2))
                x1 = min(W-1, int(cx + w/2))
                y1 = min(H-1, int(cy + h/2))
                
                new_img = img[y0:y1+1,x0:x1+1]
                new_img = cv2.resize(new_img, (W,H), \
                    interpolation=cv2.INTER_NEAREST if 'label' in key else cv2.INTER_LINEAR)

                sample[key] = new_img
            
        sample['bbox'] = np.array([0,0,W-1,H-1])

        return sample

class CenterCropFromRoI(object):
    def __init__(self,input_size,k=1.5):
        self.k = k
        self.input_size = input_size

    def __call__(self, sample):
        key_list = sample.keys()
        if not 'roi' in key_list:
            return sample

        roi = sample['roi']
        H,W = sample['image'].shape[:2]

        if np.random.rand() < 0.1:
            # sample['whole']
            roi = np.array([0,0,W-1,H-1])

        for key in key_list:
            if ('image' in key or 'label' in key) and not 'pre' in key:
                img = sample[key]
                H,W = img.shape[:2]

                x0 = max(0, roi[0])
                y0 = max(0, roi[1])
                x1 = min(W-1, roi[2])
                y1 = min(H-1, roi[3])
                new_img = img[y0:y1+1,x0:x1+1]
                # assert new_img.shape[0] > 0 and new_img.shape[1] > 0, '%d %d %d %d %d %d %d %d\n'%(x0,y0,x1,y1,W,H,img.shape[0], img.shape[1])
                new_img = cv2.resize(new_img, (W,H), \
                    interpolation=cv2.INTER_NEAREST if 'label' in key else cv2.INTER_LINEAR)

                sample[key] = new_img
        # sample['roi'] = np.array([x0, y0, x1, y1])
        # if not self.keep_roi:
        sample['roi'] = np.array([0,0,self.input_size[1],self.input_size[0]])

        return sample

class GetID(object):
    def __call__(self, sample):
        label = sample['label'].copy()
        sample['ID'] = list(set(label.flatten()))[1:]
        return sample

class GetHeatMap(object):
    def __init__(self, sigma=2.0):
        self.th = 4.6052 ## what's this??
        self.delta = math.sqrt(self.th * 2)
        self.sigma = sigma
        self.input_size = 481
        self.feat_size = 61
        self.radius = 3

    def __call__(self, sample):
        # t_kp = sample['top_keypoint']
        # l_kp = sample['left_keypoint']
        # b_kp = sample['bottom_keypoint']
        # r_kp = sample['right_keypoint']
        kp_key_words = ['top_keypoint', 'left_keypoint', 'bottom_keypoint', 'right_keypoint', 'center_keypoint']
        heat_key_words = ['top_heatmap', 'left_heatmap', 'bottom_heatmap', 'right_heatmap', 'center_heatmap']
        offset_key_words = ['top_offset', 'left_offset', 'bottom_offset', 'right_offset', 'to_del']
        weight_key_words = ['top_weight', 'left_weight', 'bottom_weight', 'right_weight', 'to_del']

        if sample['neg']:
            for h_key, os_key, w_key in zip(heat_key_words, offset_key_words, weight_key_words):
                sample[h_key] = np.zeros((self.feat_size, self.feat_size))
                sample[os_key] = np.zeros(2)
                sample[w_key] = np.zeros((self.feat_size, self.feat_size))
            
            sample.pop('to_del')
            return sample

        for key, h_key, os_key, w_key in zip(kp_key_words, heat_key_words, offset_key_words, weight_key_words):
            cx, cy = sample[key]
            cx, cy = cx*self.feat_size/self.input_size, cy*self.feat_size/self.input_size
            offset_x, offset_y = cx - int(cx), cy - int(cy)
            cx, cy = int(cx), int(cy)

            x0 = int(max(0, cx - self.delta*self.sigma+0.5))
            y0 = int(max(0, cy - self.delta*self.sigma+0.5))
            x1 = int(min(self.feat_size, cx + self.delta*self.sigma+1.5))
            y1 = int(min(self.feat_size, cy + self.delta*self.sigma+1.5))

            exp_factor = 1 / 2.0 / self.sigma / self.sigma
            heatmap = np.zeros((self.feat_size, self.feat_size))

            x_vec = (np.arange(x0, x1) - cx)**2
            y_vec = (np.arange(y0, y1) - cy)**2
            xv, yv = np.meshgrid(x_vec, y_vec)

            # arr_exp = (xv + yv) <= self.radius**2

            # arr_sum = exp_factor * np.maximum(xv + yv - self.radius**2, 0)
            arr_sum = exp_factor * (xv + yv)
            arr_exp = np.exp(-arr_sum)
            arr_exp[arr_sum > self.th] = 0
            heatmap[y0:y1, x0:x1] = arr_exp

            weight = np.zeros_like(heatmap)
            weight[cy,cx] = 1
            
            sample[h_key] = heatmap
            sample[os_key] = np.array([offset_x, offset_y])
            sample[w_key] = weight

        sample.pop('to_del')

        # img = sample['label']
        # n = img.sum()
        # h, w = img.shape
        # if n == 0:
        #     sample['center_keypoint'] = np.array([0.5, 0.5])
        #     sample['heatmap_label'] = np.zeros((self.size,self.size))
        #     return sample
        # gridx, gridy = np.meshgrid(range(w), range(h))
        # cx = (gridx*img).sum() / n / w
        # cy = (gridy*img).sum() / n / h
        # sample['center_keypoint'] = np.array([cx, cy])
        # cx,cy = int(cx*self.size), int(cy*self.size)

        # x0 = int(max(0, cx - self.delta*self.sigma+0.5))
        # y0 = int(max(0, cy - self.delta*self.sigma+0.5))
        # x1 = int(min(self.size, cx + self.delta*self.sigma+1.5))
        # y1 = int(min(self.size, cy + self.delta*self.sigma+1.5))

        # exp_factor = 1 / 2.0 / self.sigma / self.sigma
        # heatmap = np.zeros((self.size, self.size))

        # x_vec = (np.arange(x0, x1) - cx)**2
        # y_vec = (np.arange(y0, y1) - cy)**2
        # xv, yv = np.meshgrid(x_vec, y_vec)

        # arr_exp = (xv + yv) <= self.radius**2

        # # arr_sum = exp_factor * np.maximum(xv + yv - self.radius**2, 0)
        # arr_sum = exp_factor * (xv + yv)
        # arr_exp = np.exp(-arr_sum)
        # arr_exp[arr_sum > self.th] = 0
        # heatmap[y0:y1, x0:x1] = arr_exp
        # sample['heatmap_label'] = heatmap
        return sample

class GetMean(object):
    def __init__(self):
        pass

    def __call__(self, sample):
        # img = sample['image']
        # mean = sample['image'].reshape(-1,3).mean(axis=0).astype(np.int)
        # mean = tuple([int(i) for i in mean])
        mean = (127,127,127)
        sample['mean'] = mean
        return sample

class GetBoundingBox(object):
    def __init__(self, with_heat=False):
        self.with_heat = with_heat

    def __call__(self, sample):
        label = sample['label']
        y, x = np.where(label > 0)
        if len(x) < 32:
            # sample['neg'] = True
            if self.with_heat:
                sample['top_keypoint'] = np.zeros(2)
                sample['left_keypoint'] = np.zeros(2)
                sample['bottom_keypoint'] = np.zeros(2)
                sample['right_keypoint'] = np.zeros(2)
                sample['center_keypoint'] = np.zeros(2)
            sample['bbox'] = np.array([0,0,label.shape[1]-1, label.shape[0]-1])
            return sample

        ind = np.argmin(x)
        l_kp = np.array([x[ind], y[ind]])
        ind = np.argmax(x)
        r_kp = np.array([x[ind], y[ind]])
        ind = np.argmin(y)
        t_kp = np.array([x[ind], y[ind]])
        ind = np.argmax(y)
        b_kp = np.array([x[ind], y[ind]])
        ct_kp = np.array([(l_kp[0] + r_kp[0]) // 2, (t_kp[1] + b_kp[1]) // 2])
        bbox = np.array([l_kp[0], t_kp[1], r_kp[0], b_kp[1]])

        if self.with_heat:
            sample['top_keypoint'] = t_kp
            sample['left_keypoint'] = l_kp
            sample['bottom_keypoint'] = b_kp
            sample['right_keypoint'] = r_kp
            sample['center_keypoint'] = ct_kp

        sample['bbox'] = bbox

        return sample

class GetROIs(object):
    def __init__(self, shift_ratio=0.2, scale=0.1):
        self.shift_ratio = shift_ratio
        self.scale = scale
        self.ext = 1.5

    def rand_shift(self):
        if np.random.rand() < 0.1:
            return 2*(np.random.rand()-0.5)*self.shift_ratio*2
        return 2*(np.random.rand()-0.5)*self.shift_ratio
    
    def rand_scale(self):
        return 1 + 2*(np.random.rand()-0.5)*self.scale

    def __call__(self, sample):
        x0, y0, x1, y1 = sample['bbox']
        shape = sample['label'].shape
        w = x1 - x0 + 1
        h = y1 - y0 + 1
        cx = (x1 + x0) / 2
        cy = (y1 + y0) / 2
        cx += self.rand_shift()*w
        cx = max(0,min(cx, shape[1]-1))
        cy += self.rand_shift()*h
        cy = max(0,min(cy, shape[0]-1))
        w = self.rand_scale()*(w*self.ext)
        h = self.rand_scale()*(h*self.ext)
        x0 = max(0,int(cx - w/2))
        x1 = min(shape[1]-1, int(x0 + w - 1))
        y0 = max(0,int(cy - h/2))
        y1 = min(shape[0]-1, int(y0 + h - 1))
        sample['roi'] = np.array([int(x0),int(y0),int(x1),int(y1)])
        return sample

class GetVerifyBBox(object):
    def __init__(self, K=1000, num_pos=16, num_total=64, thres_high=0.7, thres_low=0.5):
        self.K = K
        self.num_pos = num_pos
        self.num_total = num_total
        self.thres_high = thres_high
        self.thres_low = thres_low

    def __call__(self, sample):
        bbox = sample['bbox']
        cx_gt = (bbox[2] + bbox[0]) // 2
        cy_gt = (bbox[3] + bbox[1]) // 2
        w_gt = bbox[2] - bbox[0]
        h_gt = bbox[3] - bbox[1]

        if self.K == 1:
            sample['verify_label'] = np.array([1], dtype=np.int32)
            sample['verify_bbox'] = bbox.copy().reshape((1,4))
            return sample

        img = sample['image']
        image_h, image_w = img.shape[:2]

        cx = np.random.rand(self.K)*image_w
        cy = np.random.rand(self.K)*image_h
        w = np.maximum(1,np.random.rand(self.K)*image_w)
        h = np.maximum(1,np.random.rand(self.K)*image_h)


        cx_ = cx_gt + 0.2*(np.random.rand(self.K)-0.5)*w_gt
        cy_ = cy_gt + 0.2*(np.random.rand(self.K)-0.5)*h_gt
        w_ = np.maximum(1,w_gt*(np.random.rand(self.K)*1.5+0.5))
        h_ = np.maximum(1,h_gt*(np.random.rand(self.K)*1.5+0.5))

        cx = np.concatenate([cx, cx_], axis=0)
        cy = np.concatenate([cy, cy_], axis=0)
        w = np.concatenate([w, w_], axis=0)
        h = np.concatenate([h, h_], axis=0)

        x0 = np.maximum(0, cx - w/2)
        y0 = np.maximum(0, cy - h/2)
        x1 = np.minimum(image_w, cx + w/2)
        y1 = np.minimum(image_h, cy + h/2)

        area1 = np.maximum(0, x1 - x0) * np.maximum(0, y1 - y0)
        area2 = np.maximum(0, bbox[2] - bbox[0]) * np.maximum(0, bbox[3] - bbox[1])
        l = np.maximum(x0, bbox[0])
        r = np.minimum(x1, bbox[2])
        t = np.maximum(y0, bbox[1])
        d = np.minimum(y1, bbox[3])
        inter = np.maximum(0, r - l) * np.maximum(0, d - t)
        iou = inter / (area1 + area2 - inter)

        pos = np.where(iou > self.thres_high)
        neg = np.where(iou < self.thres_low)


        def select(x, keep_num):
            num = x[0].shape[0]
            if  num <= keep_num:
                return x, num
                
            keep = np.arange(num)
            np.random.shuffle(keep)
            keep = keep[:keep_num]
            return tuple(y[keep] for y in x), keep_num

        pos, num_pos = select(pos, self.num_pos)
        # print(iou[pos].mean(), num_pos)
        neg, num_neg = select(neg, self.num_total - num_pos)

        label = np.zeros(self.num_total)
        label[:num_pos] = 1

        x0_p = x0[pos]
        x1_p = x1[pos]
        y0_p = y0[pos]
        y1_p = y1[pos]

        x0_n = x0[neg]
        x1_n = x1[neg]
        y0_n = y0[neg]
        y1_n = y1[neg]

        bbox_pos = np.stack([x0_p, y0_p, x1_p, y1_p], axis=-1)
        bbox_neg = np.stack([x0_n, y0_n, x1_n, y1_n], axis=-1)

        sample['verify_label'] = label
        sample['verify_bbox'] = np.concatenate([bbox_pos, bbox_neg], axis=0)

        return sample

class RandomMorphologyEx(object):
    def __init__(self, max_times=5):
        self.max_times = max_times
    
    def __call__(self, sample):
        label = sample['label']
        kernel = np.ones((5,5),np.uint8) 
        if np.random.rand() > 0.5:
            func = [cv2.erode, cv2.dilate]
        else:
            func = [cv2.dilate, cv2.erode]
        for f in func:
            t = int(np.random.rand()*self.max_times + 1)
            label = f(label,kernel,iterations = t)
        sample['label'] = label
        return sample

class RandomTranspose(object):
    def __init__(self,trans_prob=0.0):
        self.trans_prob = 1.0 - trans_prob
    def __call__(self, sample):
        prob = np.random.rand()
        step = (1.0 - self.trans_prob)/3
        count = 0
        if prob <= self.trans_prob:
            count = 0
        elif prob > self.trans_prob and prob <= self.trans_prob+step:
            count = 1
        elif prob > self.trans_prob+step and prob <= self.trans_prob+2*step:
            count = 2
        else:
            count = 3
        key_list = sample.keys()
        for key in key_list:
            if 'image' not in key and 'label' not in key:
                continue
            image = sample[key]
            image_trans = np.rot90(image, k=count)
            sample[key] = image_trans
        return sample

class ColorJitter(object):
    def __init__(self, brightness=None, contrast=None, saturation=None, *args, **kwargs):
        if not brightness is None and brightness>0:
            self.brightness = [max(1-brightness, 0), 1+brightness]
        if not contrast is None and contrast>0:
            self.contrast = [max(1-contrast, 0), 1+contrast]
        if not saturation is None and saturation>0:
            self.saturation = [max(1-saturation, 0), 1+saturation]

    def __call__(self, im_lb):
        im = im_lb['image']
        temp_im = Image.fromarray(im.astype(dtype=np.uint8),mode='RGB')
        r_brightness = random.uniform(self.brightness[0], self.brightness[1])
        r_contrast = random.uniform(self.contrast[0], self.contrast[1])
        r_saturation = random.uniform(self.saturation[0], self.saturation[1])
        temp_im = ImageEnhance.Brightness(temp_im).enhance(r_brightness)
        temp_im = ImageEnhance.Contrast(temp_im).enhance(r_contrast)
        temp_im = ImageEnhance.Color(temp_im).enhance(r_saturation)
        im = np.array(temp_im)
        im_lb['image'] = im
        return im_lb

def onehot(label, num):
    m = label
    one_hot = np.eye(num)[m]
    return one_hot

class Multiscale(object):
    def __init__(self, rate_list):
        self.rate_list = rate_list

    def __call__(self, sample):
        image = sample['image']
        row, col, _ = image.shape
        image_multiscale = []
        for rate in self.rate_list:
            rescaled_image = cv2.resize(image, None, fx=rate, fy=rate, interpolation=cv2.INTER_LINEAR)
            sample['image_%f'%rate] = rescaled_image
        return sample

class AugLastFrame(object):
    def __init__(self):
        self.mor = RandomMorphologyEx()
        
    def __call__(self, sample):
        tmp_sample = {'label': sample['last_label']}
        tmp_sample = self.mor(tmp_sample)
        sample['last_label'] = tmp_sample['label']
        return sample

class GetOptFlow(object):
    def __call__(self,sample):
        img_p = sample['last_image']
        img = sample['image']
        flow = cv2.calcOpticalFlowFarneback(img_p,img, None, 0.5, 3, 15, 3, 5, 1.2, 0)
        mask = sample['last_label']
    
    
class RandomAffine(object):
    """Randomly affine label"""
    def __init__(self, empty_mask,affine_ratio, is_optflow=True, is_continuous=False):
        self.empty_mask = empty_mask
        self.affine_ratio = affine_ratio
        self.optflow = is_optflow
        self.dis_opt_flow = cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_ULTRAFAST)
        self.seg_interpolation = cv2.INTER_LINEAR if is_continuous else cv2.INTER_NEAREST
    
    def getrand(self):
        return self.affine_ratio * (2*np.random.rand() - 1)
    
    def OpticalFlowFusion(self, prvs, next, mask):
        h,w,_ = prvs.shape
        prvs = cv2.cvtColor(prvs, cv2.COLOR_BGR2GRAY)
        next = cv2.cvtColor(next, cv2.COLOR_BGR2GRAY)
        flow = self.dis_opt_flow.calc(prvs, next, None)
        idx_x, idx_y = np.where(mask == 1)
            # new_mask = np.zeros_like(mask)
        for x,y in zip(idx_x, idx_y):
            dy, dx = flow[x,y]
            newx = min(max(0,int(x + dx)),h-1)
            newy = min(max(0,int(y + dy)),w-1)
            mask[newx, newy] = 1
            
        return mask
    
    def __call__(self, sample):
        is_empty = (np.random.rand() < self.empty_mask)
    
        if 'last_label' in sample.keys():
            label = sample['last_label']
            if is_empty:
                new_label = np.zeros_like(label)
            else:
                rows,cols = label.shape
                pts1 = np.float32([[0,0],[0,cols],[rows,0],[rows,cols]])
                pts2 = np.float32([[self.getrand()*rows,self.getrand()*cols], \
                                   [self.getrand()*rows,(1-self.getrand())*cols], \
                                   [(1-self.getrand())*rows,self.getrand()*cols], \
                                   [(1-self.getrand())*rows,(1-self.getrand())*cols]])
                M = cv2.getPerspectiveTransform(pts1,pts2)
                new_label = cv2.warpPerspective(label,M,(cols,rows),flags=self.seg_interpolation,borderValue=0)
                
                if self.optflow:
                    img = sample['image'].astype(np.uint8)
                    prvs = cv2.warpPerspective(img,M,(cols,rows),flags=self.seg_interpolation,borderValue=127)
                    new_label = self.OpticalFlowFusion(prvs, img, new_label)
    
            new_label[np.where(new_label == 255)] = 0
            sample['last_label'] = new_label
    
        if 'last_gray_image' in sample.keys():
            gray = sample['last_gray_image']
            if is_empty:
                new_gray = np.zeros_like(gray)
            else:
                new_gray = cv2.warpPerspective(gray,M,(cols,rows),flags=cv2.INTER_LINEAR,borderValue=0)
            sample['last_gray_image'] = new_gray
    
        return sample
    
class RandomHole(object):
    """Randomly crop holes in label"""
    def __init__(self, hole_p, hole_num, hole_ratio, hole_area, is_continuous=False):
        self.hole_num = hole_num
        self.hole_ratio = hole_ratio
        self.hole_area = hole_area
        self.hole_p = hole_p
        self.seg_interpolation = cv2.INTER_LINEAR if is_continuous else cv2.INTER_NEAREST
    
    def __call__(self, sample):
        if np.random.rand() > self.hole_p:
            return sample
    
        if 'last_label' in sample.keys():
            label = sample['last_label']
            rows,cols = label.shape
            hole = np.ones_like(label)
            num = int(np.random.rand()*self.hole_num + 0.5)
            for i in range(num):
                area = np.sum(label)
                area *= np.random.rand() * (self.hole_area[1] - self.hole_area[0]) + self.hole_area[0]
                r = (self.hole_ratio-1) * np.random.rand() + 1
                a = int(math.sqrt(r*area))
                b = int(math.sqrt(area/r))
                cx = int(np.random.rand()*cols)
                cy = int(np.random.rand()*rows)
                angle = int(np.random.rand()*180)
                cv2.ellipse(hole, (cx, cy), (a, b), angle, 0, 360, 0, -1)
    
            new_label = label * hole
            sample['last_label'] = new_label
    
        return sample



import torch
from torch.utils.data import Sampler
import numpy as np
import pickle as pk
from collections import Counter
from sklearn.utils import shuffle
import math
import pandas as pd


class StratifiedSampler(Sampler):
    """Stratified Sampling
    Provides equal representation of target classes in each batch
    """
    def __init__(self, 
                class_vector, 
                batch_size, 
                rpos=1, 
                rneg=4):
        '''
        args:
            rpos and rneg: this is only applied to the binary classification problem.
        '''

        self.class_vector = class_vector
        self.batch_size = batch_size

        self.rpos = rpos
        self.rneg = rneg

        if isinstance(class_vector, torch.Tensor):
            y = class_vector.cpu().numpy()
        else:
            y = np.array(class_vector)        

        y_counter = Counter(y)
        self.data = pd.DataFrame({'y': y})

        if len(y_counter.keys()) < 2:
            raise ValueError("class number must be greater or equal than 2, but got {}".format(len(y_counter.keys())))
        # only implemented for binary classification, 1:pos, 0:neg
        elif len(y_counter.keys()) == 2:
            
            ratio = (rneg, rpos)
            
            self.class_batch_size = {
            k: math.ceil(batch_size * ratio[k] / sum(ratio))
            for k in y_counter.keys()
            }

            if rpos / rneg > y_counter[1] / y_counter[0]:
                add_pos = math.ceil(rpos / rneg * y_counter[0]) - y_counter[1]

                # print("-" * 50)
                # print("To balance ratio, add %d pos imgs (with replace = True)" % add_pos)
                # # print(add_pos)
                # print("-" * 50)

                pos_samples = self.data[self.data.y == 1].sample(add_pos, replace=True)

                assert pos_samples.shape[0] == add_pos

                self.data = self.data.append(pos_samples, ignore_index=False)
            else:
                add_neg = math.ceil(rneg / rpos * y_counter[1]) - y_counter[0]

                # print("-" * 50)
                # print("To balance ratio, add %d neg imgs repeatly" % add_neg)
                # print("-" * 50)

                neg_samples = self.data[self.data.y == 0].sample(add_neg, replace=True)

                assert neg_samples.shape[0] == add_neg

                self.data = self.data.append(neg_samples, ignore_index=False)

        # print("-" * 50)
        # print("after complementary the ratio, having %d images" % self.data.shape[0])
        # print("-" * 50)
        else:
            self.class_batch_size = {
            k: math.ceil(n * batch_size / y.shape[0])
            for k, n in y_counter.items()
            }

        self.real_batch_size = int(sum(self.class_batch_size.values()))

    def gen_sample_array(self):
        # sampling for each class
        def sample_class(group):
            n = self.class_batch_size[group.name]
            return group.sample(n) if group.shape[0] >= n else group.sample(n, replace=True)

        # sampling for each batch
        data = self.data.copy()

        data['idx'] = data.index

        data = data.reset_index()

        result = []
        while True:
            try:
                batch = data.groupby('y', group_keys=False).apply(sample_class)
                assert len(
                    batch) == self.real_batch_size, 'not enough instances!'
            except (ValueError, AssertionError, AttributeError) as e:
                break
            result.extend(shuffle(batch.idx))

            data.drop(index=batch.index, inplace=True)
            # print(len(data), len(result))
        return result

    def __iter__(self):
        self.index_list = self.gen_sample_array()
        return iter(self.index_list)

    def __len__(self):
        try:
            l = len(self.index_list)
        except:
            l = len(self.class_vector)
        return l



import torch
import torch.nn as nn

'''
This file includes the pytorch implementation of partial AUC optimization, including one-way and two-way partial AUC.

Note that, this code is based on our follwoing research:

"Zhiyong Yang, Qianqian Xu, Shilong Bao, Yuan He, Xiaochun Cao and Qingming Huang. 
    When All We Need is a Piece of the Pie: A Generic Framework for Optimizing Two-way Partial AUC. ICML-2021. (Long talk)
", and

"Zhiyong Yang, Qianqian Xu, Shilong Bao, Yuan He, Xiaochun Cao and Qingming Huang. 
    Optimizing Two-way Partial AUC with an End-to-end Framework. T-PAMI'2022.

"

'''

class SquareAUCLoss(nn.Module):
    def __init__(self, 
                gamma=1,
                E_k=0,
                weight_scheme='Poly',
                num_classes=2,
                reduction='mean',
                first_state_loss=None,
                eps=1e-6,
                **kwargs):
        super(SquareAUCLoss, self).__init__()

        '''
        args:
            gamma: safe margin in square loss (default = 1.0)
            E_k: warm-up epoch (default = 0), when epoch > E_k, the partial AUC will be conducted.
            weight_scheme: weight scheme, 'Poly' or 'Exp' 
            num_classes: only support binary classification
            reduction: loss aggregated manner (default = 'mean')
            eps: use to avoid zero gradient, users can ignore this

            first_state_loss: warmup loss (default = None), could be 'SquareAUCLoss()' 
                                or other pytorch supported loss such as CE.
        '''

        self.gamma = gamma
        self.reduction = 'mean'

        if num_classes != 2:
            raise ValueError("The current version only supports binary classification!")

        self.num_classes = num_classes
        self.reduction = reduction

        # adopt re_weight func after E_k epoch....
        self.E_k = E_k

        if weight_scheme is not None and weight_scheme not in ['Poly', 'Exp']:
            raise ValueError("weight_scheme should range in [Poly, Exp]")

        self.weight_scheme = weight_scheme

        self.eps = eps

        self.first_stage_loss = first_state_loss

        if kwargs is not None:
            self.__dict__.update(kwargs)

    def _check_input(self, pred, target):
        assert pred.max() <= 1 and pred.min() >= 0
        assert target.min() >= 0
        assert pred.shape == target.shape

    def forward(self, pred, target, epoch=0, **kwargs):
        pred = pred.squeeze(-1)
        target = target.squeeze(-1)
        self._check_input(pred, target)

        if self.first_stage_loss is not None and epoch < self.E_k:
            return self.first_stage_loss(pred, target.float())

        pred_p = pred[target.eq(1)]
        pred_n = pred[target.ne(1)]

        n_plus, n_minus = len(pred_p), len(pred_n)

        weight = self.re_weight(pred_p, pred_n)
        if pred.is_cuda and not weight.is_cuda:
            weight = weight.cuda()

        pred_p = pred_p.unsqueeze(1).expand(n_plus, n_minus)
        pred_n = torch.reshape(pred_n, (1, n_minus))

        loss = weight * (self.gamma + pred_n - pred_p) ** 2
        
        return loss.mean() if self.reduction == 'mean' else loss.sum()

    def re_weight(self, pred_p, pred_n):
        return torch.ones(pred_p.shape[0], pred_n.shape[0])

class PAUCLoss(SquareAUCLoss):
    def __init__(self, gamma=1.0, 
                 E_k=0,
                 weight_scheme='Poly',
                 num_classes = 2, 
                 reduction='mean', 
                 AUC_type='OP',
                 first_state_loss=None,
                 **kwargs):

        '''
        AUC_type = OP, TP (OPAUC or TPAUC optimization)
        '''
        
        super(PAUCLoss, self).__init__(gamma, 
                                        E_k,
                                        weight_scheme,
                                        num_classes,
                                        reduction,
                                        first_state_loss,
                                        **kwargs)
        
        assert AUC_type in ['OP', 'TP'], 'AUC_type now only supports OP and TP achemes'
        
        self.AUC_type = AUC_type

        if kwargs is not None:
            self.__dict__.update(kwargs)

    def re_weight(self, pred_p, pred_n):
        '''
        return:
            must be the (len(pred_p), len(pred_n)) matrix 
                    for element-wise multiplication
        '''

        if self.weight_scheme not in ['Poly', 'Exp']:
            raise ValueError('weight_scheme 4 TPAUC must be included in [Ploy, Exp], but weight_scheme %s' % self.weight_scheme)
        
        if self.weight_scheme == 'Poly':
            col_pred_p = torch.pow((1 - pred_p + self.eps), self.gamma)
            row_pred_n = torch.pow(pred_n + self.eps, self.gamma)
        else:
            col_pred_p = 1 - torch.exp(- self.gamma * (1 - pred_p))
            row_pred_n = 1 - torch.exp(- self.gamma * pred_n)

        if self.AUC_type == 'OP':
            col_pred_p = torch.ones_like(pred_p)

        return torch.mm(col_pred_p.unsqueeze_(1), row_pred_n.unsqueeze_(0))

class MinMaxPAUC(SquareAUCLoss):
    def __init__(self, 
                gamma=1,
                E_k=0,
                weight_scheme=None,
                num_classes=2,
                eps=1e-6,
                AUC_type='OP',
                first_state_loss=None,
                reg_a=0.0,
                reg_b=0.0,
                **kwargs):
        
        super(MinMaxPAUC, self).__init__()

        '''
        args:
            reg_a and reg_b: weight of the strong convex constraint
            first_state_loss: warmup loss (default = None), could be 'SquareAUCLoss()' or other pytorch supported loss such as CE.
        '''

        self.a = torch.zeros(10, dtype=torch.float64, device="cuda", requires_grad=True)
        self.b = torch.zeros(8, dtype=torch.float64, device="cuda", requires_grad=True)

        assert self.a.requires_grad == True
        assert self.b.requires_grad == True

        self.gamma = gamma
        self.E_k = E_k

        if weight_scheme is not None and weight_scheme not in ['Poly', 'Exp']:
            raise ValueError
        self.weight_scheme = weight_scheme

        self.num_classes = num_classes

        self.AUC_type = AUC_type
        self.first_stage_loss = first_state_loss
        self.eps = eps

        self.reg_a = reg_a
        self.reg_b = reg_b

        if kwargs is not None:
            self.__dict__.update(kwargs)
        
    def function_w(self, x, y):

        assert x.shape == y.shape, "dim of x and y must be the same!"

        return 2 * x * y - torch.square(x)
    
    def function_v(self, b, a1, a2, e, f):
        return self.function_w(b, e + f) - self.function_w(a1, e) - self.function_w(a2, f)

    def forward(self, pred, target, epoch=0, **kwargs):

        if self.first_stage_loss is not None and epoch < self.E_k:
            return self.first_stage_loss(pred, target.float())
        
        pred_p = pred[target.eq(1)]
        pred_n = pred[target.ne(1)]

        
        if self.weight_scheme == 'Poly':
            v_plus = torch.pow(1 - pred_p + self.eps, self.gamma)
            v_minus = torch.pow(pred_n + self.eps, self.gamma)
        else:
            v_plus = 1 - torch.exp(- self.gamma * (1 - pred_p))
            v_minus = 1 - torch.exp(- self.gamma * pred_n)

        if self.AUC_type == 'OP':
            v_plus = torch.ones_like(pred_p)
        
        c_plus = v_plus.mean()
        c_minus = v_minus.mean()
        f_plus = (v_plus * pred_p).mean()
        f_minus = (v_minus * pred_n).mean()
        f_plus_sq = (v_plus * pred_p.square()).mean()
        f_minus_sq = (v_minus * pred_n.square()).mean()
        loss = 0.5 * self.function_v(self.b[0], self.a[0], self.a[1], c_plus, c_minus) - \
             self.function_v(self.a[2], self.b[1], self.b[2], c_minus, f_plus) + \
             self.function_v(self.b[3], self.a[3], self.a[4], c_plus, f_minus) + \
             0.5 * self.function_v(self.b[4], self.a[5], self.a[6], c_minus, f_plus_sq) + \
             0.5 * self.function_v(self.b[5], self.a[7], self.a[8], c_plus, f_minus_sq) - \
             self.function_v(self.a[9], self.b[6], self.b[7], f_plus, f_minus) 
        return loss.mean() + self.strong_convex_loss()
        
    def strong_convex_loss(self):
        return self.reg_a * self.a.square().sum() - self.reg_b * self.b.square().sum()


__copyright__ = 'Shilong Bao'
__email__ = 'baoshilong@iie.ac.cn'


import imp
from .AUROCLoss import SquareAUCLoss, ExpAUCLoss, HingeAUCLoss
from .PAUROCLoss import PAUCLoss, MinMaxPAUC


def get_losses(args):
    return eval(args.loss_type)(**(args.loss_params))



import os
import os.path as osp
import cv2
import pickle
import lmdb
from tqdm import tqdm

from .base_dataset import pil_loader


def build_lmdb(save_path, metas, commit_interval=1000):
    if not save_path.endswith('.lmdb'):
        raise ValueError("lmdb_save_path must end with 'lmdb'.")
    if osp.exists(save_path):
        print('Folder [{:s}] already exists.'.format(save_path))
        return

    if not osp.exists('/'.join(save_path.split('/')[:-1])):
        os.makedirs('/'.join(save_path.split('/')[:-1]))

    data_size_per_img = cv2.imread(metas[0][0], cv2.IMREAD_UNCHANGED).nbytes
    data_size = data_size_per_img * len(metas)
    env = lmdb.open(save_path, map_size=data_size * 10)
    txn = env.begin(write=True)

    shape = dict()

    print('Building lmdb...')
    for i in tqdm(range(len(metas))):
        image_filename = metas[i][0]
        img = pil_loader(filename=image_filename)
        assert img is not None and len(img.shape) == 3

        txn.put(image_filename.encode('ascii'), img.copy(order='C'))
        shape[image_filename] = '{:d}_{:d}_{:d}'.format(img.shape[0], img.shape[1], img.shape[2])

        if i % commit_interval == 0:
            txn.commit()
            txn = env.begin(write=True)

    pickle.dump(shape, open(os.path.join(save_path, 'meta_info.pkl'), "wb"))

    txn.commit()
    env.close()
    print('Finish writing lmdb.')

def get_all_files(dir, ext):
    for e in ext:
        if dir.lower().endswith(e):
            return [dir]

    if not osp.isdir(dir):
        return []

    file_list = os.listdir(dir)
    ret = []
    for i in file_list:
        ret += get_all_files(osp.join(dir, i), ext)
    return ret


import numpy as np
import os
import scipy.io as scio

from abc import ABC, abstractmethod
from torch.utils.data import Dataset
from torchvision import transforms
from PIL import Image, ImageFile
from imblearn.under_sampling import TomekLinks,\
    InstanceHardnessThreshold, NearMiss
from imblearn.over_sampling import ADASYN, BorderlineSMOTE

from . import custom_transforms as tr


ImageFile.LOAD_TRUNCATED_IMAGES = True
def pil_loader(filename, label=False):
    ext = (os.path.splitext(filename)[-1]).lower()
    if ext == '.png' or ext == '.jpeg' or ext == '.ppm' or ext == '.jpg':
        img = Image.open(filename)
        if not label:
            img = img.convert('RGB')
            img = np.array(img).astype(dtype=np.uint8)
            img = img[:,:,::-1]  #convert to BGR
        else:
            if img.mode != 'L' and img.mode != 'P':
                img = img.convert('L')
            img = np.array(img).astype(dtype=np.uint8)
    elif ext == '.mat':
        img = scio.loadmat(filename)
    elif ext == '.npy':
        img = np.load(filename, allow_pickle=True)
    else:
        raise NotImplementedError('Unsupported file type %s'%ext)

    return img


class BaseDataset(Dataset,ABC):
    def __init__(self, split, input_size, norm_params, resampler_type):
        super().__init__()
        self.input_size = input_size
        self.norm_params = norm_params

        if split == 'train':
            self.transform = self.transform_train()
        elif split == 'val':
            self.transform = self.transform_validation()
            resampler_type = 'None'
        elif split == 'test':
            self.transform = self.transform_validation()
            resampler_type = 'None'
        else:
            raise ValueError

        resample_dict = self.resample_dict()
        if resampler_type not in resample_dict.keys():
            raise RuntimeError('Unknown sampler_type: %s'%resampler_type)
        self.resampler = resample_dict[resampler_type]

    def resample_dict(self):
        return {
            'TL': TLResampler,
            'IHT': IHTResampler,
            'NM': NMResampler,
            'BS': BSResampler,
            'ADA': ADAResampler,
            'None': EmptyResampler
        }
    
    def resample(self, x, y):
        return self.resampler(x, y)

    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __len__(self):
        return 0

    @abstractmethod
    def __str__(self):
        pass
    
    @staticmethod
    def modify_commandline_options(parser,istrain=False):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        return parser

    def transform_train(self):
        temp = []
        temp.append(tr.Resize(self.input_size))

        temp.append(tr.RandomHorizontalFlip())
        temp.append(tr.RandomRotate(15))
        temp.append(tr.RandomCrop(self.input_size))

        temp.append(tr.Normalize(self.norm_params.mean, self.norm_params.std))
        temp.append(tr.ToTensor())
        composed_transforms = transforms.Compose(temp)
        return composed_transforms

    def transform_validation(self):
        temp = []
        temp.append(tr.Resize(self.input_size))
        # temp.append(tr.RandomCrop(self.args.input_size))
        temp.append(tr.Normalize(self.norm_params.mean, self.norm_params.std))
        temp.append(tr.ToTensor())
        composed_transforms = transforms.Compose(temp)
        return composed_transforms


class EmptyResampler():
    """
    Empty Resampler. Do nothing.
    """
    def resample(self, x, y):
        return x, y


class TLResampler():
    """
    Dataset resampled by TomekLinks
    """
    def resample(self, x, y):
        resampler = TomekLinks(sampling_strategy='majority')
        return resampler.fit_resample(x, y)


class IHTResampler():
    """
    Dataset resampled by InstanceHardnessThreshold
    """
    def resample(self, x, y):
        resampler = InstanceHardnessThreshold(sampling_strategy='majority')
        return resampler.fit_resample(x, y)


class NMResampler():
    """
    Dataset resampled by NearMiss
    """
    def resample(self, x, y):
        resampler = NearMiss(sampling_strategy='majority')
        return resampler.fit_resample(x, y)


class BSResampler():
    """
    Dataset resampled by BorderlineSMOTE
    """
    def resample(self, x, y):
        resampler = BorderlineSMOTE()
        return resampler.fit_resample(x, y)


class ADAResampler():
    """
    Dataset resampled by ADASYN
    """
    def resample(self, x, y):
        resampler = ADASYN(sampling_strategy='minority')
        return resampler.fit_resample(x, y)



import pytest
import json
import sys
sys.path.append("./")
from easydict import EasyDict as edict


@pytest.mark.skip()
def load_json(dataset=None, model=None):
    with open('./uni_test/configs/base_config.json', 'r') as f:
        args = json.load(f)
        args = edict(args)

    if dataset is not None:
        with open('./uni_test/configs/datasets/%s.json'%dataset, 'r') as f:
            args.dataset.update(edict(json.load(f)))
    
    if model is not None:
        with open('./uni_test/configs/models/%s.json'%model, 'r') as f:
            args.model.update(edict(json.load(f)))

    return args

def test_model():
    from dataloaders import get_datasets
    from dataloaders import get_data_loaders
    from models import generate_net

    for dataset in ['cifar-10-long-tail', 'tiny-imagenet-200']:
        args = load_json(dataset)
        train_set, val_set, test_set = get_datasets(args.dataset)
        train_loader, val_loader, test_loader = get_data_loaders(
            train_set,
            val_set,
            test_set,
            32,
            32
        )
        for model_type in ['mlp', 'resnet20', 'resnet18', 'densenet121']:
            args = load_json(dataset, model_type)
            if model_type == 'mlp':
                args.model.dim[0] = 3 * args.dataset.input_size[0] * args.dataset.input_size[1]
            model = generate_net(args.model).cuda()

            for i, (img, lbl) in enumerate(train_loader):
                img = img.cuda()
                lbl = lbl.cuda()
                out = model(img)
                assert out.shape == (img.shape[0], args.model.num_classes)                
                if (i+1)%4 == 0:
                    break

if __name__ == '__main__':
    test_model()



__copyright__ = 'Shilong Bao'
__email__ = 'baoshilong@iie.ac.cn'

from .minmax_opt import SGD4MinMaxPAUC

from torch.optim import SGD
from torch.optim import Adadelta
from torch.optim import Adagrad
from torch.optim import Adam
from torch.optim import AdamW




import pytest
import json
import sys
sys.path.append("./")
from easydict import EasyDict as edict
import torch


@pytest.mark.skip()
def load_json(dataset=None, loss=None):
    with open('./uni_test/configs/base_config.json', 'r') as f:
        args = json.load(f)
        args = edict(args)

    if dataset is not None:
        with open('./uni_test/configs/datasets/%s.json'%dataset, 'r') as f:
            args.dataset.update(edict(json.load(f)))
    
    if loss is not None:
        with open('./uni_test/configs/losses/%s.json'%loss, 'r') as f:
            args.training.update(edict(json.load(f)))

    return args

def test_losses_type():
    '''
        check if the loss function type is correct
    '''
    from losses import (get_losses, SquareAUCLoss, \
          HingeAUCLoss, ExpAUCLoss, PAUCLoss)

    for loss_type in ['SquareAUCLoss', 'ExpAUCLoss', 'HingeAUCLoss', 'OPAUCLoss', 'TPAUCLoss']:
        args = load_json(loss=loss_type)
        criterion = get_losses(args.training)
        assert type(criterion) == eval(args.training.loss_type)

def test_losses_decrese():
    '''
        check if the loss can drop normally
    '''
    from dataloaders import get_datasets
    from dataloaders import get_data_loaders
    from models import generate_net
    from losses import get_losses

    for dataset in ['cifar-10-long-tail', 'tiny-imagenet-200']:
        args = load_json(dataset)
        train_set, val_set, test_set = get_datasets(args.dataset)
        train_loader, val_loader, test_loader = get_data_loaders(
            train_set,
            val_set,
            test_set,
            32,
            32
        )
        for loss_type in ['SquareAUCLoss', 'ExpAUCLoss', 'HingeAUCLoss', 'OPAUCLoss', 'TPAUCLoss']:
            args = load_json(dataset, loss_type)
            model = generate_net(args.model).cuda()
            criterion = get_losses(args.training)
            optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

            train_loss = []
            for i, (img, lbl) in enumerate(train_loader):
                img = img.cuda()
                lbl = lbl.cuda()
                out = model(img)
                if out.shape[1] == 1:
                    out = torch.sigmoid(out)
                else:
                    out = torch.softmax(out, dim=1)

                loss = criterion(out, lbl)
                train_loss.append(loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                if (i+1)%100 == 0:
                    break

            assert sum(train_loss[:10]) > sum(train_loss[-10:]), train_loss

if __name__ == '__main__':
    test_losses_type()
    test_losses_decrese()



import pytest
import sys
sys.path.append("./")
import torch


@pytest.mark.skip()
def gen_data(num_classes):
    y_true = torch.randint(0, num_classes, (256, 1)).numpy()
    if num_classes == 2:
        y_pred = torch.randn((256, 1))
        y_pred = torch.sigmoid(y_pred).numpy()
    else:
        y_pred = torch.randn((256, num_classes))
        y_pred = torch.softmax(y_pred, dim=1).numpy()
    return y_true, y_pred

def test_metrics():
    from metrics import AUC
    from metrics import p2AUC

    eps = 1e-4
    for num_classes in range(3,10):
        for i in range(20):
            y_true, y_pred = gen_data(num_classes)
            # print(y_true, y_pred)
            auc_ovo = AUC(y_true, y_pred, 'ovo', acc=False)
            auc_ova = AUC(y_true, y_pred, 'ova', acc=False)
            auc_ovo_acc = AUC(y_true, y_pred, 'ovo', acc=True)
            assert abs(auc_ovo - auc_ovo_acc) < eps, \
                'auc: %.2f, auc_acc: %.2f'%(auc_ovo, auc_ovo_acc)

    for i in range(20):
        y_true, y_pred = gen_data(2)
        auc_ovo = AUC(y_true, y_pred, 'ovo', acc=False)
        pauc_03 = p2AUC(y_true, y_pred, 1, 0.3)
        pauc_06 = p2AUC(y_true, y_pred, 1, 0.6)
        pauc_10 = p2AUC(y_true, y_pred, 1, 1.0)
        p2auc_03 = p2AUC(y_true, y_pred, 0.3, 0.3)
        p2auc_06 = p2AUC(y_true, y_pred, 0.6, 0.6)

        assert abs(auc_ovo - pauc_10) < eps
        assert pauc_03 <= pauc_06 <= pauc_10
        assert p2auc_03 <= pauc_03
        assert p2auc_06 <= pauc_06
        assert p2auc_03 <= p2auc_06

if __name__ == '__main__':
    test_metrics()



from sklearn.metrics import roc_auc_score
import numpy as np
import math

def argTopk(score, k):
    return argBottomk(-score, k)

def argBottomk(score, k):
    
    return devec(np.argpartition(devec(score), k-1)[:k])

def vec(x):
    return x.reshape(-1,1)

def devec(x):
    return x.squeeze() if x.shape[0] > 1 else x 

def subIndexNegPosK(score, label, min_tpr, max_fpr):
    negIndex = np.where(label != 1)[0]
    posIndex = np.where(label == 1)[0]
    nP, nN = len(posIndex), len(negIndex)
    kpos = math.floor(nP * min_tpr)
    kneg = math.floor(nN * max_fpr)
    negTopIndex = negIndex[argTopk(score[negIndex], kneg)]
    posBotIndex = posIndex[argBottomk(score[posIndex], kpos)]

    return devec(np.vstack((vec(posBotIndex), vec(negTopIndex))))

def p2AUC(y_true, y_pred, min_tpr, max_fpr):
    subIndex = subIndexNegPosK(y_pred, y_true, min_tpr, max_fpr)
    y_true = y_true[subIndex]
    y_pred = y_pred[subIndex]
    return roc_auc_score(y_true, y_pred)



import torch
import torch.nn as nn
from abc import abstractmethod

"""
Implementation of 
        "Zhiyong Yang, Qianqian Xu, Shilong Bao, Xiaochun Cao and Qingming Huang. 
            Learning with Multiclass AUC: Theory and Algorithms. T-PAMI, 2021."
"""

class AUCLoss(nn.Module):
    
    ''' 
        args:
            num_classes: number of classes (mush include params)

            gamma: safe margin in pairwise loss (default=1.0) 

            transform: manner to compute the multi-classes AUROC Metric, either 'ovo' or 'ova' (default as 'ovo' in our paper)
    
    '''
    def __init__(self,
                 num_classes,
                 gamma=1,
                 transform='ovo', *kwargs):
        super(AUCLoss, self).__init__()

        if transform != 'ovo' and transform != 'ova':
            raise Exception("type should be either ova or ovo")
        self.num_classes = num_classes
        self.gamma = gamma
        self.transform = transform

        if kwargs is not None:
            self.__dict__.update(kwargs)
    
    def _check_input(self, pred, target):
        assert pred.max() <= 1 and pred.min() >= 0
        assert target.min() >= 0
        assert pred.shape[0] == target.shape[0]

    def forward(self, pred, target, **kwargs):
        '''
        args:
            pred: score of samples residing in [0,1]. 
            For examples, with respect to binary classification tasks, pred = torch.Sigmoid(...)
            o.w. pred = torch.Softmax(...) 

            target: index of classes. In particular, w.r.t. binary classification tasks, we regard y=1 as pos. instances.

        '''
        self._check_input(pred, target)

        if self.num_classes == 2:
            Y = target.float()
            numPos = torch.sum(Y.eq(1))
            numNeg = torch.sum(Y.eq(0))
            Di = 1.0 / numPos / numNeg
            return self.calLossPerCLass(pred.squeeze(1), Y, Di, numPos)
        else:
            if self.transform == 'ovo':
                factor = self.num_classes * (self.num_classes - 1)
            else:
                factor = 1

            Y = torch.stack(
                [target.eq(i).float() for i in range(self.num_classes)],
                1).squeeze()

            N = Y.sum(0)  
            D = 1 / N[target.squeeze().long()]  

            loss = torch.Tensor([0.]).cuda()
            if self.transform == 'ova':
                ones_vec = torch.ones_like(D).cuda()
            for i in range(self.num_classes):
                if self.transform == 'ovo':
                    Di = D / N[i]
                else:
                    fac = torch.tensor([1.0]).cuda() / (N[i] * (N.sum() - N[i]))
                    Di = fac * ones_vec
                Yi, predi = Y[:, i], pred[:, i]
                loss += self.calLossPerCLass(predi, Yi, Di, N[i])

            return loss / factor

    def calLossPerCLass(self, predi, Yi, Di, Ni):
        
        return self.calLossPerCLassNaive(predi, Yi, Di, Ni)

    @abstractmethod
    def calLossPerCLassNaive(self, predi, Yi, Di, Ni):
        pass


class SquareAUCLoss(AUCLoss):
    def __init__(self, num_classes, gamma=1, transform='ovo', **kwargs):
        super(SquareAUCLoss, self).__init__(num_classes, gamma, transform)

        # self.num_classes = num_classes
        # self.gamma = gamma

        if kwargs is not None:
            self.__dict__.update(kwargs)

    def calLossPerCLassNaive(self, predi, Yi, Di, Ni):
        diff = predi - self.gamma * Yi
        nD = Di.mul(1 - Yi)
        fac = (self.num_classes -
               1) if self.transform == 'ovo' else torch.tensor(1.0).cuda()
        S = Ni * nD + (fac * Yi / Ni)
        diff = diff.reshape((-1, ))
        S = S.reshape((-1, ))
        A = diff.mul(S).dot(diff)
        nD= nD.reshape((-1, ))
        Yi= Yi.reshape((-1, ))
        B = diff.dot(nD) * Yi.dot(diff)
        return 0.5 * A - B

class HingeAUCLoss(AUCLoss):
    def __init__(self, num_classes, gamma=1, transform='ovo', **kwargs):
        super(HingeAUCLoss, self).__init__(num_classes, gamma, transform)

        if kwargs is not None:
            self.__dict__.update(kwargs)

    def calLossPerCLassNaive(self, predi, Yi, Di, Ni):
        fac = 1 if self.transform == 'ova' else (self.num_classes - 1)
        delta1 = (fac / Ni) * Yi * predi
        delta2 = Di * (1 - Yi) * predi
        return fac * self.gamma - delta1.sum() + delta2.sum()


class ExpAUCLoss(AUCLoss):
    def __init__(self, num_classes, gamma=1, transform='ovo', **kwargs):
        super(ExpAUCLoss, self).__init__(num_classes,gamma, transform)
        
        if kwargs is not None:
            self.__dict__.update(kwargs)

    def calLossPerCLassNaive(self, predi, Yi, Di, Ni):
        C1 = Yi * torch.exp(-self.gamma * predi)
        C2 = (1 - Yi) * torch.exp(self.gamma * predi)
        C2 = Di * C2
        return C1.sum() * C2.sum()


import torch
import torch.nn as nn
import torchvision
import re

from .load import load_pretrained_model


def densenet121(args):
        model = torchvision.models.densenet121(pretrained=False)
        num_feats = model.classifier.in_features
        model.classifier = nn.Linear(num_feats, args.num_classes)

        if args.pretrained is not None:
            state_dict = torch.load(args.pretrained)
            pattern = re.compile(
        r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
            for key in list(state_dict.keys()):
                res = pattern.match(key)
                if res:
                    new_key = res.group(1) + res.group(2)
                    state_dict[new_key] = state_dict[key]
                    del state_dict[key]

            load_pretrained_model(model, state_dict)

        return model



import warnings
import numpy as np 
from sklearn.metrics import roc_auc_score

def AUC(y_true, y_pred, multi_type='ovo', acc=True):
    """
        Compute Area Under the Receiver Operating Characteristic Curve (AUROC).
        Note:
            This function can be only used with binary, multiclass AUC (either 'ova' or 'ovo').

    """

    y_true = y_true.flatten()

    if not isinstance(y_true, np.ndarray):
        warnings.warn("The type of y_ture must be np.ndarray")
        y_true = np.asarray(y_true)
    
    if not isinstance(y_pred, np.ndarray):
        warnings.warn("The type of y_pred must be np.ndarray")
        y_pred = np.asarray(y_pred)
    
    if len(np.unique(y_true)) == 2:
        assert len(y_pred) == len(y_true)
        return roc_auc_score(y_true=y_true, y_score=y_pred)
    elif multi_type == 'ova':
        return roc_auc_score(y_true=y_true, y_score=y_pred, multi_class='ovr')
    elif multi_type == 'ovo':
        if acc:
            return fast_multi_class_auc_score(y_true=y_true, y_pred=y_pred)
        else:
            auc = multi_class_auc_score(y_true=y_true, y_pred=y_pred)
        return auc
    else:
        raise NotImplementedError('multi_type must be contained in [ova, ovo]')

def multi_class_auc_score(y_true, y_pred, **kwargs):
    n = y_true.max() + 1

    def bin_auc(label, pred, i, j):
        msk1 = (label == i)
        msk2 = (label == j)
        y1 = pred[msk1, i]
        y2 = pred[msk2, i]
        return np.mean([ix > jx for ix in y1 for jx in y2])

    return np.mean([
        bin_auc(y_true, y_pred, i, j) for i in range(n) for j in range(n)
        if i != j
    ])

def fast_multi_class_auc_score(y_true, y_pred, **kwargs):
    classes = np.unique(y_true)
    num_class = len(classes)
    sum_cls = np.array([(y_true == i).sum() for i in range(num_class)])

    def bin_auc(label, pred, idx, sum_cls):
        pd = pred[:, idx]
        lbl = label
        r = np.argsort(pd)
        lbl = lbl[r]
        sum_cls = sum_cls[lbl]

        loc_idx = np.where(lbl == idx)[0][::-1]
        weight = np.zeros((len(lbl)))
        for i in range(len(loc_idx) - 1):
            weight[loc_idx[i+1] + 1:loc_idx[i]] = i + 1
        weight[:loc_idx[-1]] = len(loc_idx)

        res = (weight / sum_cls).sum() / len(loc_idx) / (num_class - 1)

        return res

    return np.mean([
        bin_auc(y_true, y_pred, idx, sum_cls) for idx in range(num_class)
    ])


from .resnet import resnet18
from .resnet_s import resnet20
from .densenet import densenet121
from .mlp import mlp

def generate_net(args):
    if args.num_classes == 2:
        args.num_classes = 1
    if not args.model_type in globals().keys():
        raise NotImplementedError("there has no %s" % (args.model_type))

    return globals()[args.model_type](args)

__all__ = ['generate_net']



from torch.utils.data import DataLoader
from .dataset import DatasetRaw, DatasetFile
from .sampler import StratifiedSampler


def get_datasets(args):
    train_set = DatasetFile(args, 'train')
    val_set = DatasetFile(args, 'test')
    test_set = DatasetFile(args, 'val')

    return train_set, val_set, test_set


def get_data_loaders(train_set,
                     val_set,
                     test_set,
                     train_batch_size,
                     test_batch_size,
                     num_workers=4,
                     rpos = 1,
                     rneg = 4):
    sampler = StratifiedSampler(train_set.get_labels(),
                                train_batch_size,
                                rpos = rpos,
                                rneg = rneg)
                                
    train_loader = DataLoader(train_set,
                              batch_size=sampler.real_batch_size,
                            #   shuffle=True,
                              sampler=sampler,
                              num_workers=num_workers)
    val_loader = DataLoader(val_set,
                            batch_size=test_batch_size,
                            shuffle=True,
                            num_workers=num_workers)
    test_loader = DataLoader(test_set,
                             batch_size=test_batch_size,
                             shuffle=True,
                             num_workers=num_workers)
    return train_loader, val_loader, test_loader


__all__ = ['DatasetRaw', 'DatasetFile', 'get_datasets', 'get_data_loaders', 'StratifiedSampler']



import pytest
import json
import sys
sys.path.append("./")
from easydict import EasyDict as edict


@pytest.mark.skip()
def load_json(dataset=None):
    with open('XCurveOpt/AUROC/uni_test/configs/base_config.json', 'r') as f:
        args = json.load(f)
        args = edict(args)

    if dataset is not None:
        with open('XCurveOpt/AUROC/uni_test/configs/datasets/%s.json'%dataset, 'r') as f:
            args.dataset.update(edict(json.load(f)))

    return args.dataset

def test_resampler():
    from XCurveOpt.AUROC.dataloaders import get_datasets
    from XCurveOpt.AUROC.dataloaders.base_dataset import (EmptyResampler, TLResampler, IHTResampler, 
            NMResampler, BSResampler, ADAResampler)

    resampler_type_2_cls = {
        'None': EmptyResampler,
        'TL': TLResampler,
        'IHT': IHTResampler,
        'NM': NMResampler,
        'BS': BSResampler,
        'ADA': ADAResampler
    }

    args = load_json()
    for rs_type in resampler_type_2_cls.keys():
        args.resampler_type = rs_type
        train_set, val_set, test_set = get_datasets(args)
        assert train_set.resampler == resampler_type_2_cls[args.resampler_type]
        assert val_set.resampler == EmptyResampler
        assert test_set.resampler == EmptyResampler

def test_single_dataset():
    from XCurveOpt.AUROC.dataloaders import get_datasets

    for dataset in ['debug_data_npy']:
        args = load_json(dataset)
        train_set, val_set, test_set = get_datasets(args)
        
        img, lbl = train_set.__getitem__(0)
        assert img.shape == (3, args.input_size[0], args.input_size[1])

def test_dataloader_binary_cls():
    from XCurveOpt.AUROC.dataloaders import get_datasets
    from XCurveOpt.AUROC.dataloaders import get_data_loaders

    for dataset in ['debug_data_npy']:
        args = load_json(dataset)
        train_set, val_set, test_set = get_datasets(args)
        train_loader, val_loader, test_loader = get_data_loaders(
            train_set,
            val_set,
            test_set,
            32,
            32
        )

        for i, (img, lbl) in enumerate(train_loader):
            assert img.shape[1:] == (3, args.input_size[0], args.input_size[1])
            assert img.shape[0] == lbl.shape[0]
            assert len(lbl.shape) == 1, lbl.shape
            assert lbl.max() == 1 and lbl.min() == 0
            if (i+1)%10 == 0:
                break

def test_dataloader_multi_cls():
    from XCurveOpt.AUROC.dataloaders import get_datasets
    from XCurveOpt.AUROC.dataloaders import get_data_loaders

    for dataset in ['debug_data_npy']:
        args = load_json(dataset)
        train_set, val_set, test_set = get_datasets(args)
        train_loader, val_loader, test_loader = get_data_loaders(
            train_set,
            val_set,
            test_set,
            32,
            32
        )

        for i, (img, lbl) in enumerate(train_loader):
            assert img.shape[1:] == (3, args.input_size[0], args.input_size[1])
            assert img.shape[0] == lbl.shape[0]
            assert len(lbl.shape) == 1
            assert lbl.max() == max(list(args.class2id.values())) and lbl.min() == 0
            if (i+1)%10 == 0:
                break

if __name__ == '__main__':
    test_resampler()
    test_single_dataset()
    test_dataloader_binary_cls()
    test_dataloader_multi_cls()



import setuptools
import os
if os.path.exists('README.md'):
    with open('README.md', encoding='utf-8') as fp:
        long_description = fp.read()
else:
    long_description = ' '
if os.path.exists('requirements.txt'):
    with open('requirements.txt', encoding='utf-8') as fp:
        install_requires = fp.read()
        install_requires = install_requires.split('\n')
else:
    install_requires = []

setuptools.setup(
    name="XCurve",                     # This is the name of the package
    version="0.0.5",                      # The initial release version
    author="qqgroup",                     # Full name of the author
    author_email='  ',
    description="machine learning package",
    url='https://gitee.com/Statusrank/xcurve-opt',
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),    # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.8',                # Minimum version requirement of the package
    py_modules=["XCurve"],             # Name of the python package
    package_dir={'XCurve':'XCurve'},
    install_requires=install_requires       # Install other dependencies if any
)



__copyright__ = 'Shilong Bao'
__email__ = 'baoshilong@iie.ac.cn'

import torch
from torch.optim.optimizer import Optimizer 

class SGD4MinMaxPAUC(Optimizer):

    def __init__(self, 
                params, 
                a = None, 
                b = None, 
                lr = 1e-2,
                momentum=0.,
                dampening=0.,
                nesterov=False,
                weight_decay = 0., 
                clip_value = 5.0,
                epoch_to_opt=10,
                **kwargs):
        
        assert a is not None, 'Found no variable a!'
        assert b is not None, 'Found no variable b!'

        '''
        The SGD optimization for Min-Max Partial AUC, including OPAUC and TPAUC. 

        args:
            params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups

            lr (float): learning rate
            weight_decay (float): regularizer weights
            clip_value (float): to clip gradient
            epoch_to_opt: 
            a (tensor): reformulation weights a
            b (tensor): reformulation weights b
        
        The other args stay the same with SGD in pytorch
        '''

        if lr <= 0.0:
            raise ValueError('Invalid Learning rate {} - should be >= 0.'.format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0:
            raise ValueError('Invalid weight_decay {} - should be >= 0.'.format(weight_decay))
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        
        self.eopch_to_opt = epoch_to_opt
        self.a = a
        self.b = b 

        assert self.a.requires_grad == True
        assert self.b.requires_grad == True

        defaults = dict(lr=lr,
                        weight_decay=weight_decay,
                        a = self.a, 
                        b = self.b,
                        clip_value = clip_value,
                        momentum=momentum,
                        dampening=dampening,
                        nesterov=nesterov 
                        )
        
        super(SGD4MinMaxPAUC, self).__init__(params=params, 
                                        defaults=defaults)


    def zero_grad(self, set_to_none: bool = False):
        r"""Sets the gradients of all optimized :class:`torch.Tensor` s to zero.

        Arguments:
            set_to_none (bool): instead of setting to zero, set the grads to None.
                This is will in general have lower memory footprint, and can modestly improve performance.
                However, it changes certain behaviors. For example:
                1. When the user tries to access a gradient and perform manual ops on it,
                a None attribute or a Tensor full of 0s will behave differently.
                2. If the user requests ``zero_grad(set_to_none=True)`` followed by a backward pass, ``.grad``\ s
                are guaranteed to be None for params that did not receive a gradient.
                3. ``torch.optim`` optimizers have a different behavior if the gradient is 0 or None
                (in one case it does the step with a gradient of 0 and in the other it skips
                the step altogether).
        """
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is not None:
                    if set_to_none:
                        p.grad = None
                    else:
                        if p.grad.grad_fn is not None:
                            p.grad.detach_()
                        else:
                            p.grad.requires_grad_(False)
                        p.grad.zero_()
        
        if self.a.grad is not None and self.b.grad is not None:
            self.a.grad.zero_()
            self.b.grad.zero_()
        else: 
            self.a.grad = None
            self.b.grad = None

    def __setstate__(self, state):
        super(SGD4MinMaxPAUC, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)
    
    @torch.no_grad()
    def step(self, epoch=0):
        """Performs a single optimization step.
        
        args:
            epoch (int): the current epoch of optimization
        """
        for group in self.param_groups:
            
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            lr = group['lr']

            clip_value = group['clip_value']    
            
            for p in group['params']:
                if p.grad is None:
                    continue

                d_p = p.grad

                if d_p.is_sparse:
                    raise RuntimeError('does not support sparse gradients !!!')
                
                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)
                
                if momentum != 0:
                    param_state = self.state[p]

                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf 
                
                # clip the gradient
                d_p = torch.clamp(d_p, -clip_value, clip_value)
                p.data.add_(d_p, alpha=-lr)
            
            if epoch >= self.eopch_to_opt:
                a, b = group['a'], group['b']
                
                if a.grad is None or b.grad is None:
                    raise ValueError('the grad of a/b is None !')
                
                d_a, d_b = a.grad, b.grad

                if d_a.is_sparse or d_b.is_sparse: 
                    raise RuntimeError('does not support sparse gradients !!!')

                if momentum != 0:
                    param_state_a = self.state[a]
                    param_state_b = self.state[b]

                    if 'momentum_buffer' not in param_state_a:
                        buf_a = param_state_a['momentum_buffer'] = torch.clone(d_a).detach()
                        buf_b = param_state_b['momentum_buffer'] = torch.clone(d_b).detach()
                    else:
                        buf_a = param_state_a['momentum_buffer']
                        buf_a.mul_(momentum).add_(d_a, alpha=1 - dampening)

                        buf_b = param_state_b['momentum_buffer']
                        buf_b.mul_(momentum).add_(d_b, alpha=1 - dampening)
                    if nesterov:
                        d_a = d_a.add(buf_a, alpha = momentum)
                        d_b = d_b.add(buf_b, alpha = momentum)
                    else:
                        d_a = buf_a 
                        d_b = buf_b 

                # clip the gradient
                d_a = torch.clamp(d_a, -clip_value, clip_value)
                d_b = torch.clamp(d_b, -clip_value, clip_value)

                a.data.add_(d_a, alpha = -lr)
                b.data.add_(d_b, alpha = lr)


'''
Properly implemented ResNet-s for CIFAR10 as described in paper [1].
The implementation and structure of this file is hugely influenced by [2]
which is implemented for ImageNet and doesn't have option A for identity.
Moreover, most of the implementations on the web is copy-paste from
torchvision's resnet and has wrong number of params.
Proper ResNet-s for CIFAR10 (for fair comparision and etc.) has following
number of layers and parameters:
name      | layers | params
ResNet20  |    20  | 0.27M
ResNet32  |    32  | 0.46M
ResNet44  |    44  | 0.66M
ResNet56  |    56  | 0.85M
ResNet110 |   110  |  1.7M
ResNet1202|  1202  | 19.4m
which this implementation indeed has.
Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
[2] https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
If you use this implementation in you work, please don't forget to mention the
author, Yerlan Idelbayev.
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
import math

from torch.autograd import Variable

__all__ = ['ResNet', 'resnet20', 'resnet32', 'resnet44', 'resnet56', 'resnet110', 'resnet1202']

def _weights_init(m):
    classname = m.__class__.__name__

    # kaiming weight normal after default init
    if isinstance(m, nn.Conv2d):
        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
        m.weight.data.normal_(0, math.sqrt(2. / n))
    elif isinstance(m, nn.BatchNorm2d):
        m.weight.data.fill_(1)
        m.bias.data.zero_()
    elif isinstance(m, nn.Linear):
        init.kaiming_uniform_(m.weight)

class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, option='A'):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            if option == 'A':
                """
                For CIFAR10 ResNet paper uses option A.
                """
                self.shortcut = LambdaLayer(lambda x:
                                            F.pad(x[:, :, ::2, ::2], (0, 0, 0, 0, planes//4, planes//4), "constant", 0))
            elif option == 'B':
                self.shortcut = nn.Sequential(
                     nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                     nn.BatchNorm2d(self.expansion * planes)
                )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=1):
        super(ResNet, self).__init__()
        self.in_planes = 16

        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 64, num_blocks[2], stride=2)
        self.linear = nn.Linear(64, num_classes)

        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def resnet20(args):
    model = ResNet(BasicBlock, [3, 3, 3], args.num_classes)
    return model

def resnet32():
    return ResNet(BasicBlock, [5, 5, 5])


def resnet44():
    return ResNet(BasicBlock, [7, 7, 7])


def resnet56():
    return ResNet(BasicBlock, [9, 9, 9])


def resnet110():
    return ResNet(BasicBlock, [18, 18, 18])


def resnet1202():
    return ResNet(BasicBlock, [200, 200, 200])


def test(net):
    import numpy as np
    total_params = 0

    for x in filter(lambda p: p.requires_grad, net.parameters()):
        total_params += np.prod(x.data.numpy().shape)
    print("Total number of params", total_params)
    print("Total layers", len(list(filter(lambda p: p.requires_grad and len(p.data.size())>1, net.parameters()))))


if __name__ == "__main__":
    for net_name in __all__:
        if net_name.startswith('resnet'):
            print(net_name)
            test(globals()[net_name]())
            print()


import numpy as np
import os
import os.path as osp
import pandas as pd
import pickle
import lmdb

from .base_dataset import BaseDataset
from .base_dataset import pil_loader
from .utils import build_lmdb, get_all_files


class DatasetRaw(BaseDataset):
    def __init__(self, data, targets, 
            split='train',
            input_size=224,
            norm_params={'mean': [123.675, 116.280, 103.530], 
                    'std': [58.395, 57.120, 57.375]},
            resampler_type='None',
            load_data_from_file=False,
            lmdb_dir=None,
            **kwargs
        ):
        super().__init__(split, input_size, norm_params, resampler_type)
        assert targets.min() >= 0 and targets.max() > 0
        # print(targets.max(), targets.min(), len(targets), targets.sum())
        self.targets = targets

        ## a list of meta data, each element is (feature, target) or (img path, target)
        self.metas = [(d,t) for d,t in zip(data, targets)]

        ## directory path to save lmdb files
        if lmdb_dir is not None:
            lmdb_dir = osp.join(lmdb_dir, split + '.lmdb')
        self._lmdb_dir = lmdb_dir
        ## number of meta data
        self._num = len(self.metas)
        ## number of positive classes
        self._n_cls_p = int(targets.max())

        # ## info for resampling
        # if self._n_cls_p == 1:
        #     self._labels = [int(i[1].sum() != 0) for i in self.metas]
        # else:
        #     self._labels = [np.argmax(i[1]) for i in self.metas]
        # print(self._labels)

        self._cls_num_list = pd.Series(targets).value_counts().sort_index().values
        self._freq_info = [
            num * 1.0 / sum(self._cls_num_list) for num in self._cls_num_list
        ]

        self._num_classes = len(self._cls_num_list)
        # self._class_dim = len(set(self._labels))
        self._class_dim = len(set(targets))

        if load_data_from_file:
            self._getitem = self._getitem_file
            if self._lmdb_dir:
                build_lmdb(self._lmdb_dir, self.metas)
                self.initialized = False
                self._load_image = self._load_image_lmdb
            else:
                self._load_image = self._load_image_pil            
        else:
            self._getitem = self._getitem_data

    def _init_lmdb(self):
        if not self.initialized:
            env = lmdb.open(self._lmdb_dir, readonly=True, lock=False, readahead=False, meminit=False)
            self.lmdb_txn = env.begin(write=False)
            self.meta_info = pickle.load(open(os.path.join(self._lmdb_dir, 'meta_info.pkl'), "rb"))
            self.initialized = True

    def _load_image_lmdb(self, img_filename):
        self._init_lmdb()
        img_buff = self.lmdb_txn.get(img_filename.encode('ascii'))
        C, H, W = [int(i) for i in self.meta_info[img_filename].split('_')]
        img = np.frombuffer(img_buff, dtype=np.uint8).reshape(C, H, W)
        return img

    def _load_image_pil(self, img_filename):
        return pil_loader(img_filename)

    def get_class_dim(self):
        return self._class_dim

    def get_labels(self):
        return self.targets

    def get_cls_num_list(self):
        return self._cls_num_list

    def get_freq_info(self):
        return self._freq_info

    def get_num_classes(self):
        return self._num_classes

    def __len__(self):
        return self._num

    def __str__(self):
        return 'dataset: split=' + str(self.split)

    def _getitem_file(self, idx):
        sample = {
            'image': self._load_image(self.metas[idx][0]),
            'label': self.metas[idx][1]
        }
        sample = self.transform(sample)
        return sample['image'], sample['label']

    def _getitem_data(self, idx):
        sample = {
            'image': self.metas[idx][0],
            'label': self.metas[idx][1]
        }
        sample = self.transform(sample)
        return sample['image'], sample['label']
    
    def __getitem__(self, idx):
        return self._getitem(idx)


class DatasetFile(DatasetRaw):
    def __init__(self, args, split='train'):
        data_dir = osp.join(args.data_dir, split)

        ## get classes
        if not 'class2id' in args.keys():
            class2id = dict()
            for i in range(args.num_classes):
                class2id[str(i)] = i
        else:
            class2id = args.get('class2id')

        ## check file type
        if osp.isdir(data_dir):
            npy_style = False
        elif osp.exists(data_dir + '.npy'):
            data_dir += '.npy'
            npy_style = True
        else:
            raise IOError('data directory not exists: %s'%data_dir)

        ## load meta data
        data = None
        targets = None
        if npy_style:
            tmp = np.load(data_dir, allow_pickle=True).item()
            ori_data = tmp['data']
            ori_targets = tmp['targets']
            data = []
            targets = []
            for i in range(len(ori_data)):
                id_ = class2id.get(str(int(ori_targets[i])), 0)
                if id_ >= 0:
                    data.append(ori_data[i])
                    targets.append(id_)
            data = np.array(data)
            targets = np.array(targets)
            assert len(data) == len(targets)
        else:
            img_list = get_all_files(data_dir, ['jpg', 'jpeg', 'png'])
            data, targets = self._gen_metas(img_list, class2id)

        args['data'] = data
        args['targets'] = targets
        args['load_data_from_file'] = not npy_style
        args['split'] = split
        super().__init__(**args)

    def _gen_metas(self, img_list, class2id):
        data, targets = [], []
        for i in img_list:
            cls_id = class2id.get(i.split('/')[-2], 0)
            if cls_id < 0:
                continue
            data.append(i)
            targets.append(cls_id)
        targets = np.array(targets)
        return data, targets



import torch


class UnitBallClipper(object):
    def __init__(self, frequency, radius):
        self.frequency = frequency
        self.radius = radius

    def __call__(self, module):
        """
        filter the variables to get the ones you want
        """
        if hasattr(module, 'weight'):
            w = module.weight.data
            scales = torch.norm(w, 2)
            if scales > self.radius:
                w.div_(torch.norm(w, 2).expand_as(w))
                w.mul_(self.radius)


class MultiLayerPerceptron(torch.nn.Module):
    def __init__(self, dim=[21527, 128, 64, 12], num_classes=1):
        super(MultiLayerPerceptron, self).__init__()
        layers = []
        for i in range(1, len(dim)):
            layers += [
                torch.nn.Linear(dim[i-1], dim[i]),
                torch.nn.BatchNorm1d(dim[i]),
                torch.nn.ReLU()
            ]
        layers += [
            torch.nn.Linear(dim[-1], num_classes),
            torch.nn.Softmax()
        ]
        self.layers = torch.nn.Sequential(*layers)

    def forward(self, x):
        x = x.view(x.shape[0], -1)
        return self.layers(x)

def mlp(args):
    model = MultiLayerPerceptron(args.dim, args.num_classes)
    return model

# def get_model(input_dim, output_dim):
#     return torch.nn.Sequential(
#         # torch.nn.Linear(21527, 128),
#         torch.nn.Linear(input_dim, 128),
#         torch.nn.BatchNorm1d(128),
#         torch.nn.ReLU(),
#         torch.nn.Linear(128, 64),
#         torch.nn.BatchNorm1d(64),
#         torch.nn.ReLU(),
#         torch.nn.Linear(64, output_dim),
#         # torch.nn.Linear(64, 12),
#         torch.nn.Softmax()).cuda()


# def get_clipper(frequency, radius):
#     return UnitBallClipper(frequency, radius)



import os
import torch


# def load_checkpoint(checkpoint_path=None):
#     assert(checkpoint_path is not None)
#     if not os.path.isfile(checkpoint_path):
#         logger.debug("=> no checkpoint found at '{}'" .format(checkpoint_path))
#         raise RuntimeError("=> no checkpoint found at '{}'" .format(checkpoint_path))
#     checkpoint = torch.load(checkpoint_path, map_location='cpu')
#     start_it = 0
#     stage = 0
#     optimizer = None

#     if 'optimizer' in checkpoint.keys():
#         start_it = checkpoint['start_it']
#         stage = checkpoint['stage']
#         optimizer = checkpoint['optimizer']
#         state_dict = checkpoint['state_dict']
#         state_dict = {k.replace('module.','',1): v for k,v in state_dict.items()}
#     else:
#         logger.info('this checkpoint has no optimizer')
#         state_dict = {k.replace('module.','',1): v for k,v in checkpoint.items()}
#     logger.info('epoch %.06d'%(stage, start_it))

#     return state_dict, optimizer, start_it, stage

def load_pretrained_model(model, state_dict):
    model_state = model.state_dict()
    model_params = len(model_state.keys())
    checkpoint_params = len(state_dict.keys())
    # logger.info('this model has {} params; this checkpoint has {} params'.format(model_params,checkpoint_params))
    print('this model has {} params; this checkpoint has {} params'.format(model_params,checkpoint_params))
    if model_params > checkpoint_params:
        for i,param in model_state.items():
            if i not in state_dict.keys() and not i.endswith('num_batches_tracked'):
                # logger.info('this param of the model dont in the checkpoint: {} ,required grad: {}'.format(i,str(param.requires_grad)))
                print('this param of the model dont in the checkpoint: {} ,required grad: {}'.format(i,str(param.requires_grad)))
    num = 0
    total = 0
    for k,v in state_dict.items():
        total += 1
        if k in model_state.keys():
            if not isinstance(v,bool):
                if (v.size() != model_state[k].size()):
                    # logger.info('this param {} of the checkpoint dont match the model in size: '.format(k) + str(v.size()) + ' ' + str(model_state[k].size()))
                    print('this param {} of the checkpoint dont match the model in size: '.format(k) + str(v.size()) + ' ' + str(model_state[k].size()))
                    continue
            model_state[k] = v
            num += 1
        else:
            # logger.info('this param of the checkpoint dont in the model: {}'.format(k))
            print('this param of the checkpoint dont in the model: {}'.format(k))
    model.load_state_dict(model_state, strict=False)
    # logger.info('success for loading pretrained model params {}/{}!'.format(str(num), str(total)))
    print('success for loading pretrained model params {}/{}!'.format(str(num), str(total)))
